/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		modified 2005 ... 2012 by Harald Foerster		   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "moz.h"
#include "bswap32.h"
#include "file.h"
#include "nls.h"

/*
#include <sys/types.h>
#include <sys/stat.h>
#include "nscache.h"
#include "stringbuf.h"
#include <gtk/gtk.h>
#include <time.h>
*/


#define MOZ_BITMAP_SIZE		4096
#define MOZ_MAX_BLOCK_SIZE	4096

#ifdef MOZ_ENABLE_MAX_VERSION
#define MOZ_VERSION(v)		(((v) >= MOZ_MIN_VERSION) && ((v) <= MOZ_MAX_VERSION))
#else
#define MOZ_VERSION(v)		((v) >= MOZ_MIN_VERSION)
#endif

#define MOZ_LOC_INITED(v)	 ((v) & 0x80000000)
#define MOZ_LOC_FILE_INDEX(v)	(((v) & 0x30000000) >> 28)
#define MOZ_LOC_FILE_SIZE(v) 	 ((v) & 0x00FFFF00)
#define MOZ_LOC_BLOCK_COUNT(v)	(((v) & 0x03000000) >> 24)
#define MOZ_LOC_BLOCK_NUMBER(v)  ((v) & 0x00FFFFFF)
#define MOZ_LOC_GENERATION(v)	 ((v) & 0xFF)

#define MOZ_GET_MAX_BLOCK_NUMBER(bitmap_size)  	(((bitmap_size) * 8) - 1)

#define MOZ_GET_BITMAP_SIZE_1v19(file_index) \
		((131072 / 8) >> (((file_index) - 1) * 2))

#define MOZ_GET_BLOCK_OFFSET(moz_file, block_number) \
		(((moz_file)->block_size * (block_number)) + (moz_file)->bitmap_size)

/* includes terminating null byte */
#define MOZ_FILENAME_LENGTH	(14)

#define kBuckets		(1 << 5)
#define kBucketStart		(sizeof(MozCacheMapHeader))
#define kBucketStart_1v6	(sizeof(MozCacheMapHeader_1v6))

#define GetBucketIndex(hashnr)	((hashnr) & (kBuckets - 1))


typedef struct
{
	guint32	version;
	gint32	size;
	gint32	nentries;
	guint32	dirty;

} MozMapFileHeader;

typedef struct
{
	MozMapFileHeader	filehead;
	guint32			erank[kBuckets];
	guint8			reserved[MOZ_BITMAP_SIZE -
					sizeof(MozMapFileHeader) -
					(sizeof(guint32) * kBuckets)];
} MozCacheMapHeader;

typedef struct
{
	MozMapFileHeader	filehead;
	gint32			record_count;
	guint32			erank[kBuckets];
	guint32			bucket_usage[kBuckets];

} MozCacheMapHeader_1v6;

typedef struct
{
	guint32	hashnr;
	guint32	erank;
	guint32	data_loc;
	guint32	meta_loc;

} MozCacheMapRecord;

typedef struct
{
	guint32	version;	/* usefull for stand-alone metadata files */
	guint32	meta_loc;	/* for verification */
	gint32	fetch_count;
	guint32	last_fetched;
	guint32	last_modified;
	guint32	expiration;
	guint32	data_size;
	guint32	key_size;	/* includes terminating null byte */
	guint32	meta_size;	/* includes terminating null byte */
	guint8	key_start[1];	/* start of key data */

} MozCacheEntryRecord;

#define CMR_BYTESWAP_UNKNOWN	((ByteSwap32Func) -1)

enum
{
	MOZ_MAP_FILE,
	MOZ_B1_FILE,
	MOZ_B2_FILE,
	MOZ_B3_FILE,
	MOZ_FIRST_FILE	= MOZ_MAP_FILE,
	MOZ_LAST_FILE	= MOZ_B3_FILE
};

typedef struct
{
	int		fd;
	char		*block_alloc;
	size_t		alloc_size;
	unsigned int	max_block_nr;
	size_t		bitmap_size;
	const size_t	block_size;
	const char	*name;

} MozCacheFile;

typedef struct
{
	guint32			version_orig;
	guint32			version_swap;
	guint32			dirty;
	guint32			bucket_start;
	guint32			recs_per_bucket;
	ByteSwap32Func		cmr_byteswap;
	ByteSwap32Func		block_alloc_byteswap;
	int			err_map;
	int			err_block;
	int			open_mode;
	char			*clr_block;
	MozCacheMapRecord	*record_array;

} MozMapFile;

typedef struct
{
	MozCacheFile	cache_file[4];
	MozMapFile	map_file;

} MozDatabase;

static MozDatabase moz_database =
{
	/* cache_file[4] */
	{
		{ -1, NULL, 0, 0, 0,     0, "_CACHE_MAP_" },
		{ -1, NULL, 0, 0, 0,   256, "_CACHE_001_" },
		{ -1, NULL, 0, 0, 0,  1024, "_CACHE_002_" },
		{ -1, NULL, 0, 0, 0,  4096, "_CACHE_003_" }
	},

	/* map_file */
	{
		0
	}
};

static const char moz_err_cache_in_use[] =
			gettext_nop("WARNING: cache in use");

static const char moz_err_corrupted_cache_map_file[] =
			gettext_nop("Corrupted cache map file");

static void	*moz_read_db_init(char **dbname, NScacheReadNextFunc *func);
static GSList	*moz_read_db_free(void *ptr);
static int	moz_cache_check(const char *dbname);

#ifndef CACHE_READONLY
static int	moz_record_delete(nscache_record_t *);
#endif

typedef struct
{
	GSList			*slist;
	int			buckets;
	int			bucket_size;
	int			idx;

} moz_read_data_t;

NSC_BROWSER moz_browser =
{
	moz_read_db_init,
	moz_read_db_free,

#ifndef CACHE_READONLY
	moz_record_delete,
#endif
	moz_cache_check,

	gettext_nop("Mozilla")
};



static int moz_close_db(void)
{
	int		error;
	MozCacheFile	*moz_file;
	MozDatabase	*moz_db;

	moz_db = &moz_database;
	g_free((moz_db->map_file).record_array);
	(moz_db->map_file).record_array = NULL;

	error = 0;
	moz_file = &(moz_db->cache_file[MOZ_LAST_FILE]);

	do
	{
		if (moz_file->fd != -1)
		{
			if (moz_file->block_alloc)
			{
#ifndef CACHE_READONLY
				if (moz_file->alloc_size)
				{
					if (lseek(moz_file->fd, 0, SEEK_SET) != 0 ||
						write(moz_file->fd, moz_file->block_alloc,
							moz_file->alloc_size) != moz_file->alloc_size)
					{
						if (error == 0) error = errno;
					}

					moz_file->alloc_size = 0;
				}
#endif /* ! CACHE_READONLY */
				g_free(moz_file->block_alloc);
				moz_file->block_alloc = NULL;
			}

			close(moz_file->fd);
			moz_file->fd = -1;
		}
	}
	while (--moz_file >= &moz_database.cache_file[MOZ_FIRST_FILE]);

	return error;

} /* static int moz_close_db(void) */

static MozDatabase *moz_open_db(const char *dbname, MozCacheFile *moz_file, int mode, int errmesg)
{
	int			dirty;
	const char		*str;
	MozDatabase		*moz_db;
	MozCacheMapHeader_1v6	header, *hptr;

	moz_db = &moz_database;

	if (moz_file == NULL)
	{
		moz_file = &moz_db->cache_file[MOZ_LAST_FILE];
	}

	do
	{
		moz_file->fd = file_open(moz_file->name, mode);

		if (moz_file->fd == -1)
		{
			if (errmesg)
				gui_strerror(nscache_err_open_index_file);

			moz_close_db();

			return NULL;
		}

		moz_file->bitmap_size	= MOZ_BITMAP_SIZE;
		moz_file->max_block_nr	= MOZ_GET_MAX_BLOCK_NUMBER(MOZ_BITMAP_SIZE);
	}
	while (--moz_file >= &moz_db->cache_file[MOZ_FIRST_FILE]);

	dirty = 0;

	(moz_db->map_file).open_mode = mode;

	hptr = &header;

#ifndef CACHE_READONLY

	if (mode == FILE_OPEN_READ_WRITE)
	{
		hptr = g_malloc(sizeof(MozCacheMapHeader_1v6));
		(moz_db->cache_file[MOZ_MAP_FILE]).block_alloc = (char *) hptr;
	}
#endif

	str = NULL;

	if (read((moz_db->cache_file[MOZ_MAP_FILE]).fd, hptr, sizeof(header)) != sizeof(header))
	{
		str = gettext_nop("Error reading cache information");
	}

	else
	{
		guint32	version;

		version = (hptr->filehead).version;

		if ((moz_db->map_file).version_orig == 0)
		{
			(moz_db->map_file).version_orig = version;
			(moz_db->map_file).version_swap = UINT32_FROM_BIG_ENDIAN(version);

			if ( ! MOZ_VERSION((moz_db->map_file).version_swap) )
			{
				if (errmesg)
				{
					gui_errfmt("%s : version %d.%d",
						gettext("Mozilla cache map file not recognized"),
						(unsigned int) ((moz_db->map_file).version_swap >> 16),
						(unsigned int) ((moz_db->map_file).version_swap & 0x0000FFFF));
				}

				moz_close_db();

				return NULL;
			}
		}

		dirty = (hptr->filehead).dirty;
		(moz_db->map_file).dirty = dirty;

		if (version != (moz_db->map_file).version_orig)
		{
			str = gettext_nop("Version number differs from previous");
		}

		else if ((moz_db->map_file).version_swap < 0x00010006)
		{
			(moz_db->map_file).bucket_start    = kBucketStart;
			(moz_db->map_file).recs_per_bucket = 256;
		}

		else
		{
			guint32	hdr_recs;

			if ((moz_db->map_file).version_swap >= 0x00010013)
			{
				unsigned int file_index = MOZ_B3_FILE;

				do
				{
					moz_file = &moz_db->cache_file[file_index];

					moz_file->bitmap_size =
						MOZ_GET_BITMAP_SIZE_1v19(file_index);

					moz_file->max_block_nr =
						MOZ_GET_MAX_BLOCK_NUMBER(moz_file->bitmap_size);
				}
				while (--file_index >= MOZ_B1_FILE);
			}

			hdr_recs = UINT32_FROM_BIG_ENDIAN(hptr->record_count);

			if (hdr_recs % (kBuckets * sizeof(MozCacheMapRecord)) == 0)
			{
				(moz_db->map_file).recs_per_bucket =
							hdr_recs / kBuckets;

				(moz_db->map_file).bucket_start = kBucketStart_1v6;
			}

			else
			{
				str = gettext_nop("Unknown number of max records");
			}
		}
	}

	if (str == NULL)
	{
		static const char err_lseek[] = gettext_nop("Error executing 'lseek'");

		int length = lseek((moz_db->cache_file[MOZ_MAP_FILE]).fd, 0, SEEK_END);

		if (length > 0)
		{
			guint32 bucket_start, recs_per_bucket;

			bucket_start	= (moz_db->map_file).bucket_start;
			recs_per_bucket	= (moz_db->map_file).recs_per_bucket;

			if (length < (bucket_start + kBuckets *
				recs_per_bucket * sizeof(MozCacheMapRecord)))
			{
				str = moz_err_corrupted_cache_map_file;
			}

			else
			{
				(moz_db->map_file).record_array =
					g_malloc((recs_per_bucket + 1) *
						sizeof(MozCacheMapRecord));

				(moz_db->map_file).record_array[recs_per_bucket].hashnr = 0;

				if (lseek((moz_db->cache_file[MOZ_MAP_FILE]).fd, bucket_start, SEEK_SET) == bucket_start)
				{
					return moz_db;
				}

				str = err_lseek;
			}
		}

		else
		{
			str = err_lseek;
		}
	}

	if (errmesg)
	{
		str = gettext(str);

		if (dirty)
		{
			gui_errfmt("%s - %s",
				gettext(moz_err_cache_in_use), str);
		}

		else
		{
			gui_err(str);
		}
	}

	moz_close_db();

	return NULL;

} /* static int moz_open_db(const char*, MozCacheFile*, int, int) */

static int moz_generate_filename(MozMapFile *map_file, MozCacheMapRecord *cmr, int type, char* fname)
{
	guint32 generation =
		MOZ_LOC_GENERATION(type == 'm' ? cmr->meta_loc : cmr->data_loc);

	if (map_file->version_swap >= 0x00010013)
	{
		/* local files stored in a folder and a sub folder */

		return sprintf(fname, "%01X/%02X/%05X%c%02X",
			(cmr->hashnr & 0xF0000000) >> (7 * 4),
				(cmr->hashnr & 0x0FF00000) >> (5 * 4),
					cmr->hashnr & 0x000FFFFF,
						(char) type, generation);
	}

	return sprintf(fname, "%08X%c%02X", cmr->hashnr, (char) type, generation);
}

static moz_record_t *moz_compose_record(MozDatabase *moz_db, MozCacheMapRecord *cmr,
						MozCacheEntryRecord *entry, size_t size)
{
	moz_record_t	*rec;
	char		*key, *str, *end;
	unsigned int	index;
	guint32		len;

	if (sizeof(MozCacheEntryRecord) > size)
	{
		return NULL;
	}

	end = &((char *) entry)[size];
	key = entry->key_start;

	if (key <= (char *) entry || key > end)
	{
		return NULL;
	}

	key += UINT32_FROM_BIG_ENDIAN(entry->key_size);

	if (key <= (char *) entry || key > end)
	{
		return NULL;
	}

	len = UINT32_FROM_BIG_ENDIAN(entry->meta_size);

	if (&key[len] <= (char *) entry || &key[len] > end)
	{
		return NULL;
	}

	if (entry->version != (moz_db->map_file).version_orig)
	{
		return NULL;
	}

	/* initialize to '\0' */
	rec = g_malloc0(sizeof(moz_record_t));

	rec->recordid = cmr->hashnr;
	rec->record.content_length = UINT32_FROM_BIG_ENDIAN(entry->data_size);

	index = MOZ_LOC_FILE_INDEX(cmr->data_loc);

	if (index)
	{
		guint32		blocknr;
		MozCacheFile	*moz_file;

		moz_file = &moz_db->cache_file[index];
		blocknr  = MOZ_LOC_BLOCK_NUMBER(cmr->data_loc);

		if (blocknr > moz_file->max_block_nr)
		{
			g_free(rec);
			return NULL;
		}

		rec->record.content_offset =
			MOZ_GET_BLOCK_OFFSET(moz_file, blocknr);

		rec->record.filename = (char *) moz_file->name;
	}

	else
	{
		char fname[MOZ_FILENAME_LENGTH];
#if 0
		rec->record.content_offset = 0;
#endif
		moz_generate_filename(&(moz_db->map_file), cmr, 'd', fname);
		rec->record.filename = g_strdup(fname);
	}

	str = strchr(entry->key_start, ':');

	if (str && str[1] != '\0')
	{
		rec->record.urlstr = g_strdup(&str[1]);

		rec->record.last_accessed = UINT32_FROM_BIG_ENDIAN(entry->last_fetched);
		rec->record.last_modified = UINT32_FROM_BIG_ENDIAN(entry->last_modified);
		rec->record.expires = UINT32_FROM_BIG_ENDIAN(entry->expiration);

		str = key;
		key = tl_mem_find_str(key, len, "response-head", strlen("response-head"));

		if (key)
		{
			time_t tmp;

			key += strlen("response-head") + 1;
			len -= key - str;

			/* Content-Type: */
			str = tl_get_mime_value_str(key, len, "content-type:",
							sizeof("content-type:") - 1);
			rec->record.content_type = content_type_into_glist(str);

			/* "charset" */
			str = tl_get_mime_value_str(key, len, "charset",
							sizeof("charset") - 1);
			rec->record.charset = content_misc_into_gslist(str);

			/* Content-Encoding: */
			str = tl_get_mime_value_str(key, len, "content-encoding:",
							sizeof("content-encoding:") - 1);
			rec->record.content_encoding = content_misc_into_gslist(str);

			/* Date: */
			str = tl_get_mime_value_str(key, len, "date:", sizeof("date:") - 1);
			tmp = time_from_string(str);

			if (tmp)
			{
				rec->record.last_accessed = tmp;
			}

			/* Last-Modified: */
			str = tl_get_mime_value_str(key, len, "last-modified:",
							sizeof("last-modified:") - 1);
			tmp = time_from_string(str);

			if (tmp)
			{
				rec->record.last_modified = tmp;
			}

			/* Expires: */
			str = tl_get_mime_value_str(key, len, "expires:", sizeof("expires:") - 1);
			tmp = time_from_string(str);

			if (tmp)
			{
				rec->record.expires = tmp;
			}

		} /* if (key) */

		return rec;

	} /* if (str && str[1] != '\0') */

	nscache_record_free(&rec->record);

	return NULL;

} /* static moz_record_t *moz_compose_record(MozDatabase*, MozCacheMapRecord*,
							MozCacheEntryRecord*, size_t) */

static MozCacheEntryRecord *moz_load_entry_record(MozDatabase *moz_db,
							MozCacheMapRecord *cmr, size_t *size)
{
	unsigned int	index;
	guint32		blocknr;
	MozCacheFile	*moz_file;

	index = MOZ_LOC_FILE_INDEX(cmr->meta_loc);

	if (index == 0)
	{
		nscache_record_t	nscrec;
		char			fname[MOZ_FILENAME_LENGTH];

		/* read entry record stored in stand alone file */

		moz_generate_filename(&(moz_db->map_file), cmr, 'm', fname);

		nscrec.filename		= fname;
		nscrec.content_encoding	= NULL;
		nscrec.content_offset	= 0;

		return file_read(&nscrec, size);
	}

	moz_file = &moz_db->cache_file[index];
	blocknr  = MOZ_LOC_BLOCK_NUMBER(cmr->meta_loc);

	if (blocknr <= moz_file->max_block_nr)
	{
		unsigned int offset = MOZ_GET_BLOCK_OFFSET(moz_file, blocknr);

		if (lseek(moz_file->fd, offset, SEEK_SET) == offset)
		{
			guint32			block_count;
			MozCacheEntryRecord	*entry;

			block_count = MOZ_LOC_BLOCK_COUNT(cmr->meta_loc) + 1;

			*size	= moz_file->block_size * block_count;
			entry	= malloc(*size + 1); /* use malloc() here - like file_read() */

			if (entry)
			{
				int sz = read(moz_file->fd, entry, *size);

				if (sz != -1 && (sz == *size || /* EOF ??? */
					lseek(moz_file->fd, 0, SEEK_CUR) == offset + sz))
				{
					((char *) entry)[sz] = '\0';
					*size = sz;

					return entry;
				}

				free(entry);
			}
		}

	} /* if (blocknr <= moz_file->max_block_nr) */

	return NULL;

} /* static MozCacheEntryRecord *moz_load_entry_record(MozDatabase*, MozCacheMapRecord*, size_t*) */

static void moz_map_record_byteswap(MozCacheMapRecord *cmr)
{
	cmr->hashnr	= BSWAP32_BE_LE(cmr->hashnr);
	/*cmr->erank	= BSWAP32_BE_LE(cmr->erank);*/
	cmr->data_loc	= BSWAP32_BE_LE(cmr->data_loc);
	cmr->meta_loc	= BSWAP32_BE_LE(cmr->meta_loc);
}

static moz_record_t *moz_get_entry_info(MozDatabase *moz_db, MozCacheMapRecord *cmr)
{
	static const ByteSwap32Func swapfunc_tab[] =
	{
		BSWAP32_BIG_ENDIAN_FUNC,
		BSWAP32_LITTLE_ENDIAN_FUNC,
		CMR_BYTESWAP_UNKNOWN
	};

	const ByteSwap32Func *swapfunc_ptr = swapfunc_tab;

	for (;;)
	{
		ByteSwap32Func		swap;
		MozCacheMapRecord	cmrrec, *cmrptr;

		swap = (moz_db->map_file).cmr_byteswap;

		if (swap == CMR_BYTESWAP_UNKNOWN)
		{
			swap = *(swapfunc_ptr++);

			if (swap == CMR_BYTESWAP_UNKNOWN)
			{
				return NULL;
			}

			cmrrec = *cmr;
			cmrptr = &cmrrec;
		}

		else
		{
			cmrptr = cmr;
		}

		if (swap)
		{
			moz_map_record_byteswap(cmrptr);
		}

		if ( ! MOZ_LOC_INITED(cmrptr->data_loc) ||
				! MOZ_LOC_INITED(cmrptr->meta_loc))
		{
			if ((moz_db->map_file).cmr_byteswap != CMR_BYTESWAP_UNKNOWN)
			{
				return NULL;
			}
		}

		else
		{
			size_t			size;
			MozCacheEntryRecord	*entry;

			entry = moz_load_entry_record(moz_db, cmrptr, &size);

			if (entry)
			{
				moz_record_t *rv = moz_compose_record(moz_db, cmrptr, entry, size);

				free(entry); /* malloc'd */

				if (rv)
				{
					(moz_db->map_file).cmr_byteswap = swap;
					return rv;
				}
			}

			if ((moz_db->map_file).cmr_byteswap != CMR_BYTESWAP_UNKNOWN)
			{
				break;
			}
		}

	} /* for (;;) */

	(moz_db->map_file).err_block++;

	return NULL;

} /* static moz_record_t *moz_get_entry_info(MozDatabase *moz_db, MozCacheMapRecord *cmr) */

static int moz_read_db_next(void *ptr)
{
	MozDatabase		*moz_db;
	MozCacheMapRecord	*record_array;
	moz_record_t		*nr;
	moz_read_data_t		*data;

	data		= ptr;
	moz_db		= &moz_database;
	record_array	= (moz_db->map_file).record_array;

	while (record_array[data->idx].hashnr == 0)
	{
		int size;

		if (--data->buckets < 0)
		{
			return FALSE;
		}

		data->idx = 0;

		size = read(moz_db->cache_file[MOZ_MAP_FILE].fd,
					record_array, data->bucket_size);

		if (size != data->bucket_size)
		{
			(moz_db->map_file).err_map = 1;
			return FALSE;
		}
	}

	nr = moz_get_entry_info(moz_db, &(record_array[data->idx]));

	if (nr)
	{
		data->slist = g_slist_prepend(data->slist, nr);
	}

	data->idx++;

	return TRUE;

} /* static int moz_read_db_next(void *ptr) */

static void *moz_read_db_init(char **dbname, NScacheReadNextFunc *func)
{
	MozDatabase	*moz_db;
	moz_read_data_t	*data;

	*func = (NScacheReadNextFunc) moz_read_db_next;

	g_free(*dbname);
	*dbname = NULL;

	moz_db = &moz_database;
	(moz_db->map_file).version_orig	= 0;
	(moz_db->map_file).err_map 	= 0;
	(moz_db->map_file).err_block 	= 0;
	(moz_db->map_file).cmr_byteswap	= CMR_BYTESWAP_UNKNOWN;

	moz_db = moz_open_db(*dbname, NULL, FILE_OPEN_READ_ONLY, TRUE);

	if (moz_db == NULL)
	{
		return moz_db;
	}

	*dbname = g_strdup((moz_db->cache_file[MOZ_MAP_FILE]).name);

	nscache_db.version = g_strdup_printf("%hu.%hu",
			(unsigned short) ((moz_db->map_file).version_swap >> 16),
			(unsigned short) ((moz_db->map_file).version_swap & 0x0000FFFF));

	if ((moz_db->map_file).version_swap < 0x0001000B)
	{
		/* 8-bit => reverse byte order */
		(moz_db->map_file).block_alloc_byteswap = BSWAP32_LITTLE_ENDIAN_FUNC;
	}

	else
	{
		/* 32bit => big endian */
		(moz_db->map_file).block_alloc_byteswap = BSWAP32_BIG_ENDIAN_FUNC;
	}

	data = g_malloc(sizeof(moz_read_data_t));

	data->slist		= NULL;
	data->idx		= (moz_db->map_file).recs_per_bucket;
	data->buckets		= kBuckets;
	data->bucket_size	= data->idx * sizeof(MozCacheMapRecord);

	return data;

} /* static void *moz_read_db_init(char**, NScacheReadNextFunc*) */

static GSList *moz_read_db_free(void *ptr)
{
	int		idx;
	GSList		*retv;
	MozDatabase	*moz_db;
	moz_read_data_t	*data;
	const char	*errstr[3];

	data = ptr;
	retv = data->slist;
	g_free(data);

	moz_close_db();

	errstr[0] = errstr[1] = errstr[2] = "";

	idx	= -1;
	moz_db	= &moz_database;

	if ((moz_db->map_file).dirty)
	{
		idx = 0;
		errstr[0] = gettext(moz_err_cache_in_use);
	}

	if ((moz_db->map_file).err_block)
	{
		idx += 3; /* 2 or 3 */
		errstr[1] = gettext("Corrupted blockfile");
	}

	if ((moz_db->map_file).err_map)
	{
		idx++; /* 0; 1; 3 or 4 */
		errstr[2] = gettext(moz_err_corrupted_cache_map_file);
	}

	if (idx >= 0)
	{
		static const char *errfmtstr[] =
		{
			"%s%s",
			"%s - %s",
			"%s%s (%d)%s",
			"%s - %s (%d)%s",
			"%s - %s (%d) - %s",
			"%s%s (%d) - %s"
		};

		if ((moz_db->map_file).err_block)
		{
			if (idx == 3 && *(errstr[0]) == '\0')
			{
				idx = 5;
			}

			gui_errfmt(errfmtstr[idx], errstr[0],
				errstr[1], (moz_db->map_file).err_block, errstr[2]);
		}

		else
		{
			gui_errfmt(errfmtstr[idx], errstr[0], errstr[2]);
		}
	}

	return retv;

} /* static GSList *moz_read_db_free(void *ptr) */

static int moz_read_bucket(guint32 recordid, MozCacheMapRecord *record_array)
{
	unsigned int	bucket, bucket_size, offset;
	MozDatabase	*moz_db;

	moz_db = &moz_database;

	if (moz_db->cache_file[MOZ_MAP_FILE].fd == -1)
	{
		goto _bucket_err;
	}

	bucket_size	= (moz_db->map_file).recs_per_bucket * sizeof(MozCacheMapRecord);
	bucket		= GetBucketIndex(recordid);

	offset  = (moz_db->map_file).bucket_start;
	offset += bucket * bucket_size;

	if (lseek((moz_db->cache_file[MOZ_MAP_FILE]).fd, offset, SEEK_SET) == offset)
	{
		int sz = read((moz_db->cache_file[MOZ_MAP_FILE]).fd,
							record_array, bucket_size);
		if (sz == bucket_size)
		{
			int i;

			if ((moz_db->map_file).cmr_byteswap)
			{
				recordid = (*(moz_db->map_file).cmr_byteswap)(recordid);
			}

			record_array[(moz_db->map_file).recs_per_bucket].hashnr = 0;

			i = 0;

			for (;;)
			{
				guint32 hashnr = record_array[i].hashnr;

				if (hashnr == 0)
				{
					break;
				}

				if (hashnr == recordid)
				{
#ifndef CACHE_READONLY
					if ((moz_db->map_file).open_mode == FILE_OPEN_READ_WRITE &&
						lseek((moz_db->cache_file[MOZ_MAP_FILE]).fd,
									offset, SEEK_SET) != offset)
					{
						break;
					}
#endif /* ! CACHE_READONLY */
					return i;
				}
				
				i++;
			}

		} /* if (sz == bucket_size) */

	} /* if (lseek((moz_db->cache_file[MOZ_MAP_FILE]).fd, ... ) */

_bucket_err:
	errno = ENODATA;

	return NSC_ERR_HASHNR;

} /* static int moz_read_bucket(int, int, guint32, MozCacheMapRecord*) */

static int moz_toggle_bitmap(MozDatabase *moz_db, guint32 location, int rw_blk)
{
	unsigned int	index, offset;
	guint32		blocknr, bit_shift, block_count, mask;
	guint32		result, block_alloc, *blk_ptr;
	MozCacheFile	*moz_file;
	ByteSwap32Func	swap;

	if ( ! MOZ_LOC_INITED(location) )
	{
		return NSC_ERR_ENTRY;
	}

	index = MOZ_LOC_FILE_INDEX(location);

	if (index == 0)
	{
		return index;
	}

	moz_file = &moz_db->cache_file[index];

	if (moz_file->fd == -1)
	{
		int fd = file_open(moz_file->name, (moz_db->map_file).open_mode);

		if (fd == -1)
		{
			return NSC_ERR_BLKFILE;
		}

		moz_file->fd = fd;

		if (rw_blk & 0x01)
		{
			moz_file->block_alloc = g_malloc(moz_file->bitmap_size);

			if (read(fd, moz_file->block_alloc, moz_file->bitmap_size) != moz_file->bitmap_size)
			{
				memset(moz_file->block_alloc, '\0', moz_file->bitmap_size);
			}
		}

	} /* if (moz_file->fd == -1) */

	blocknr = MOZ_LOC_BLOCK_NUMBER(location);

	if (blocknr > moz_file->max_block_nr)
	{
		return NSC_ERR_ENTRY;
	}

	offset = (blocknr / 8) & ~(0x00000003);

	if (moz_file->block_alloc == NULL)
	{
		if ( lseek(moz_file->fd, offset, SEEK_SET) != offset ||
			read(moz_file->fd, &block_alloc, sizeof(guint32)) != sizeof(guint32) )
		{
			return NSC_ERR_ENTRY;
		}

		blk_ptr = &block_alloc;
	}

	else
	{
		blk_ptr = (guint32 *) &moz_file->block_alloc[offset];
	}

	block_count	= MOZ_LOC_BLOCK_COUNT(location) + 1;
	bit_shift	= blocknr % 32;

	if (block_count + bit_shift > 32)
	{
		return NSC_ERR_ENTRY;
	}

	mask   = (1 << block_count) - 1;
	mask <<= bit_shift;

	swap = (moz_db->map_file).block_alloc_byteswap;

	if (swap)
	{
		mask = (*swap)(mask);
	}

	result = *blk_ptr ^ mask;

	if ((result & mask) == 0)
	{

#ifndef CACHE_READONLY

		if (rw_blk == 0x03)
		{
			*blk_ptr = result;
		}
#endif
		return index;
	}

	return NSC_ERR_ENTRY;

} /* static int moz_toggle_bitmap(MozDatabase*, guint32, int) */

#ifndef CACHE_READONLY

static gint32 moz_cache_free_blocks(MozDatabase *moz_db, MozCacheMapRecord *cmr, int type)
{
	int		index;
	unsigned int	offset;
	gint32		blocknr, block_count, location, size;
	MozCacheFile	*moz_file;

	location = type == 'm' ? cmr->meta_loc : cmr->data_loc;

	index = moz_toggle_bitmap(moz_db, location, 0x03);

	if (index < 0)
	{
		return index;
	}

	if (index == 0)
	{
		/* remove meta or data file */

		guint32	rv;
		char	fname[MOZ_FILENAME_LENGTH];

		moz_generate_filename(&(moz_db->map_file), cmr, type, fname);

		if (file_remove(fname) != 0)
		{
			return type == 'm' ? NSC_ERR_ENTRY : NSC_ERR_FILE;
		}

		/* 1v19 => 0/12/34567xyy */

		if (fname[4] == '/')
		{
			/* remove sub folder */

			fname[4] = '\0';
			file_rmdir(fname);
		}

		/* result in KB, shift offset is eight */
		rv = MOZ_LOC_FILE_SIZE(location);

		return ((moz_db->map_file).version_swap < 0x0001000B) ?
			(rv << 2) :	/* KB * 4 => pages of 256 Bytes */
			(rv >> 8);	/* KB (starting at LSB) */
	}

	/* clear meta or data blocks */

	moz_file = &moz_db->cache_file[index];

	/* write allocation bits to disk */
	moz_file->alloc_size = moz_file->bitmap_size;

	block_count	= MOZ_LOC_BLOCK_COUNT(location) + 1;
	size		= moz_file->block_size * block_count;

	/* 'max_block_nr' was tested in 'moz_toggle_bitmap' */
	blocknr		= MOZ_LOC_BLOCK_NUMBER(location);
	offset		= MOZ_GET_BLOCK_OFFSET(moz_file, blocknr);

	if (lseek(moz_file->fd, offset, SEEK_SET) == offset)
	{
		gint32 sz, len;

		sz = size;

		if (size > MOZ_MAX_BLOCK_SIZE)
		{
			sz = MOZ_MAX_BLOCK_SIZE;
		}

		len = 0;

		do
		{
			if (write(moz_file->fd, (moz_db->map_file).clr_block, sz) != sz)
			{
				break;
			}

			len += sz;
		}
		while (len < size);
	}

	return ((moz_db->map_file).version_swap < 0x0001000B) ?
		(size) :		/* pages of 256 Bytes */
		((size + 1023) >> 10);	/* round up to nearest KB (starting at LSB) */

} /* static gint32 moz_cache_free_blocks(MozDatabase*, MozCacheMapRecord*, int) */

static int moz_record_delete(nscache_record_t *nsc_rec)
{
	int		i, bucket, size;
	gint32		rv, nentries, total_size;
	guint32		erank, *bucket_erank;
	MozDatabase	*moz_db;
	ByteSwap32Func	swap;

	MozCacheMapRecord	*cmr, *record_array;
	MozCacheMapHeader_1v6	*header;

	moz_record_t *moz_rec = (moz_record_t *) nsc_rec;

	moz_db		= &moz_database;
	record_array	= (moz_db->map_file).record_array;

	i = moz_read_bucket(moz_rec->recordid, record_array);

	if (i < 0)
	{
		return i;
	}

	cmr = &record_array[i];

	if ((moz_db->map_file).cmr_byteswap)
	{
		moz_map_record_byteswap(cmr);
	}

	/* test blocks */

	if ((rv = moz_toggle_bitmap(moz_db, cmr->meta_loc, 0x01)) < 0 ||
			(rv = moz_toggle_bitmap(moz_db, cmr->data_loc, 0x01)) < 0)
	{
		if (rv == NSC_ERR_ENTRY)
		{
			errno = ENODATA;
		}

		return rv;
	}

	/* first free data blocks */

	rv = moz_cache_free_blocks(moz_db, cmr, 'd');

	if (rv < 0)
	{
		return rv;
	}

	total_size = rv;

	/* then free meta blocks */

	rv = moz_cache_free_blocks(moz_db, cmr, 'm');

	if (rv < 0)
	{
		return rv;
	}

	total_size += rv;

	/* find last record */

	do
	{
		i++;
	}
	while (record_array[i].hashnr);

	/* fill the gap with the last record */

	i--;

	if (cmr != &record_array[i])
	{
		*cmr =  record_array[i];
		 cmr = &record_array[i];
	}

	/* clear last record */

	cmr->hashnr 	= 0;
	cmr->erank	= 0;
	cmr->data_loc	= 0;
	cmr->meta_loc	= 0;

	/* update eviction rank */

	erank	= 0;
	swap	= (moz_db->map_file).cmr_byteswap;

	while (--cmr >= record_array)
	{
		guint32 tmp = cmr->erank;

		if (swap)
		{
			tmp = (*swap)(tmp);
		}

		if (erank < tmp)
		{
			erank = tmp;
		}
	}

	size = (i + 1) * sizeof(MozCacheMapRecord);

	if (write((moz_db->cache_file[MOZ_MAP_FILE]).fd, record_array, size) != size)
	{
		return NSC_ERR_ENTRY;
	}

	/* write header to disk */
	(moz_db->cache_file[MOZ_MAP_FILE]).alloc_size = sizeof(MozCacheMapHeader_1v6);

	header = (MozCacheMapHeader_1v6 *)
			(moz_db->cache_file[MOZ_MAP_FILE]).block_alloc;

	/* update cache size */

	total_size = (header->filehead).size - total_size;

	if (total_size < 0)
	{
		total_size = 0;
	}

	(header->filehead).size = total_size;

	/* update number of entries */

	nentries = (header->filehead).nentries - 1;

	if (nentries < 0)
	{
		nentries = 0;
	}

	(header->filehead).nentries = nentries;

	bucket = GetBucketIndex(moz_rec->recordid);

	if ((moz_db->map_file).version_swap < 0x00010006)
	{
		bucket_erank = &((MozCacheMapHeader *) header)->erank[bucket];
	}

	else
	{
		/* update number of records in bucket */

		guint32 bcount;

		bcount	= header->bucket_usage[bucket];
		swap	= NULL;

		if ((bcount & (guint32) 0x000001FF) == 0)
		{
			swap	= BSWAP32_BE_LE_FUNC;
			i	= (*swap)(i);
		}

		header->bucket_usage[bucket] = i;

		bucket_erank = &header->erank[bucket];
	}

	*bucket_erank = (swap == NULL) ? erank : (*swap)(erank);

	return 0;

} /* static int moz_record_delete(nscache_record_t *nsc_rec) */

int moz_blk_delete_open(void)
{
	MozDatabase		*moz_db;
	MozCacheMapHeader_1v6	*header;

	moz_db = moz_open_db(nscache_db.name,
			&moz_database.cache_file[MOZ_MAP_FILE], FILE_OPEN_READ_WRITE, TRUE);

	if (moz_db == NULL)
	{
		return NSC_ERR_SILENT;
	}

	(moz_db->map_file).clr_block = g_malloc0(MOZ_MAX_BLOCK_SIZE);

	header = (MozCacheMapHeader_1v6 *)
			(moz_db->cache_file[MOZ_MAP_FILE]).block_alloc;

	(header->filehead).size =
			UINT32_FROM_BIG_ENDIAN((header->filehead).size);

	(header->filehead).nentries =
			UINT32_FROM_BIG_ENDIAN((header->filehead).nentries);

	return 0;

} /* int moz_blk_delete_open(void) */

char *moz_blk_delete_close(void)
{
	int			error;
	MozDatabase		*moz_db;
	MozCacheMapHeader_1v6	*header;

	moz_db = &moz_database;

	g_free((moz_db->map_file).clr_block);
	(moz_db->map_file).clr_block = NULL;

	header = (MozCacheMapHeader_1v6 *)
			(moz_db->cache_file[MOZ_MAP_FILE]).block_alloc;

	(header->filehead).size =
			UINT32_TO_BIG_ENDIAN((header->filehead).size);

	(header->filehead).nentries =
			UINT32_TO_BIG_ENDIAN((header->filehead).nentries);

	error = moz_close_db();

	if (error)
	{
		errno = error;
		return gettext_nop("Error writing block alloc table");
	}

	return NULL;

} /* char *moz_blk_delete_close(void) */

#endif /* ! CACHE_READONLY */

int moz_blk_read_open(void)
{
	MozDatabase *moz_db;

	moz_db = moz_open_db(nscache_db.name,
			&moz_database.cache_file[MOZ_MAP_FILE],
					FILE_OPEN_READ_ONLY, FALSE);
	if (moz_db == NULL)
	{
		return NSC_ERR_SILENT;
	}

	return 0;

} /* int moz_blk_read_open(void) */

void moz_blk_read_close(void)
{
	moz_close_db();

} /* void moz_blk_read_close(void) */

static int moz_block_access(const nscache_record_t *nsc_rec, int rw_blk)
{
	int			rv;
	moz_record_t		*moz_rec;
	MozDatabase		*moz_db;
	MozCacheMapRecord	*cmr, *record_array;

	moz_rec = (moz_record_t *) nsc_rec;

	moz_db		= &moz_database;
	record_array	= (moz_db->map_file).record_array;

	rv = moz_read_bucket(moz_rec->recordid, record_array);

	if (rv < 0)
	{
		return rv;
	}

	cmr = &record_array[rv];

	if ((moz_db->map_file).cmr_byteswap)
	{
		moz_map_record_byteswap(cmr);
	}

	if ((rv = moz_toggle_bitmap(moz_db, cmr->meta_loc, rw_blk)) < 0 ||
			(rv = moz_toggle_bitmap(moz_db, cmr->data_loc, rw_blk)) < 0)
	{
		if (rv == NSC_ERR_ENTRY)
		{
			errno = ENODATA;
		}
	}

	return rv;

} /* static int moz_block_access(const nscache_record_t, int) */

int moz_blk_access(const nscache_record_t *nsc_rec)
{
	int index = moz_block_access(nsc_rec, 0x01);

	if (index > 0)
	{
		return moz_database.cache_file[index].fd;
	}

	return -1;

} /* int moz_blk_access(const nscache_record_t *nsc_rec) */

int moz_file_access(const nscache_record_t *nsc_rec)
{
	int		dirty, rv;
	MozDatabase	*moz_db;

	dirty	= (moz_database.map_file).dirty;
	moz_db	= moz_open_db(nscache_db.name,
			&moz_database.cache_file[MOZ_MAP_FILE],
					FILE_OPEN_READ_ONLY, FALSE);

	if (moz_db == NULL)
	{
		errno = ENODATA;
		return NSC_ERR_SILENT;
	}

	rv = moz_block_access(nsc_rec, 0x00);

	moz_close_db();

	if (rv > 0)
	{
		rv = FILE_ACCESS_READ;
	}

	if (dirty == 0 && (moz_database.map_file).dirty)
	{
		gui_err(gettext(moz_err_cache_in_use));
	}

	return rv;

} /* int moz_file_access(const nscache_record_t *nsc_rec) */

static int moz_cache_check(const char *dbname)
{
	if (NSCACHE_RT_MOZILLA != NSCACHE_RT_LAST)
	{
		if (moz_open_db(dbname, NULL,
			FILE_OPEN_READ_ONLY, FALSE) == NULL)
		{
			return NSC_ERR_SILENT;
		}

		moz_close_db();
	}

	return 0;

} /* static int moz_cache_check(const char *dbname) */

/* EOF */
